/*
 * The contents of this file are subject to the Mozilla Public License 
 * Version 1.0 (the "License"); you may not use this file except in compliance 
 * with the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License. 
 * 
 * The Original Code is the roamd Version 0.1 code,
 * released on or about December 4, 1998.
 * 
 * The Initial Developer of the Original Code is John Richard Stracke,
 * Jr.  Portions created by John Richard Stracke, Jr. are Copyright (C)
 * John Richard Stracke, Jr. All Rights Reserved. 
 * 
 * Contributor(s): ______________________________________. 
 */
/*
 * Implements the password functions declared in passwd.h.
 *
 * $Revision: 1.7 $
 * $Author: francis $
 * $Date: 1998/12/17 20:07:54 $
 *
 * $Log: passwd.c,v $
 * Revision 1.7  1998/12/17 20:07:54  francis
 * Disabling the debug in de64(), it being a security risk
 *
 * Revision 1.6  1998/12/13 06:19:56  francis
 * Adding rudimentary platform awareness to the Makefile.
 * Creating crypt.h, where the platform-appropriate header file for crypt()
 * is #included.
 * Note that the platform awareness breaks non-gmake support.  I need to do
 * something about that.
 *
 * Revision 1.5  1998/12/06 20:02:21  francis
 * Adding ident lines
 *
 * Revision 1.4  1998/12/04 22:03:20  francis
 * Adding license terms to the source files.
 * Adding tryCrypt and tryCrypt.o to the clean target.
 * Deleting tryCrypt from the repository.
 * Adding Index.html and license.html.  Linking to license.html from
 * readme.html.  A bit of cosmetic work in readme.html.
 *
 * Revision 1.3  1998/12/04 20:50:12  francis
 * Fixing password authentication to make sure that the username in the
 * credentials is the owner of the requested file.
 *
 * Revision 1.2  1998/12/04 20:01:03  francis
 * Adding documentation headers and comments
 *
 */

#include "ident.h"
IDENTIFY(ROAMD_PASSWD_C,"$Id: passwd.c,v 1.7 1998/12/17 20:07:54 francis Exp $")

#include "crypt.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include <string.h>

#include <sys/types.h>
#include <pwd.h>

#include "passwd.h"

/*
 * Function to decode base64 text (needed by Basic authentication).
 */
char* de64(const char* base64)
{
  static const char digits[65]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
  static int inverseDigits[256];
  static int inverseDigitsInited=0;
  static struct {
    char* data;
    int len;
  } buff={0,0};

  if (!base64)
    {
      fprintf(stderr,"de64(): base64==NULL\n");
      errno=EFAULT;
      return 0;
    }

  if (!inverseDigitsInited)
    {
      int i;

      for (i=0; i<(sizeof(inverseDigits)/sizeof(inverseDigits[0])); i++)
        inverseDigits[i]=-1;

      for (i=0; i<((sizeof(digits)/sizeof(digits[0]))-1); i++)
        inverseDigits[digits[i]]=i;

      inverseDigitsInited=1;
    }

  if (strlen(base64)%4)
    {
      fprintf(stderr,"de64(): strlen(\"%s\")==%d\n",base64,strlen(base64));
      errno=EINVAL;
      return 0;
    }

  {
    int lenNeeded=(strlen(base64)*3)/4+1;
    if (lenNeeded>buff.len)
      {
        char* newData=realloc(buff.data,lenNeeded);
        if (!newData)
          return 0;

        buff.data=newData;
        buff.len=lenNeeded;
      }

    {
      char* curData=buff.data;
      char* maxData=buff.data+buff.len;
      const char* curBase64=base64;

      while (   ((curData+2)<maxData)
             && (curBase64)
             )
        {
          int curBase64Out[4];
          {
            int i;
            for (i=0; i<4; i++)
              curBase64Out[i]=inverseDigits[curBase64[i]];
          }

          if (curBase64[2]=='=')
            {
              curData[0]=(  (curBase64Out[0]<<2)
                          | (curBase64Out[1]>>4)
                          );
              curData[1]=0;
            }
          else
            {
              if (curBase64[3]=='=')
                {
                  curData[0]=(  (curBase64Out[0]<<2)
                              | (curBase64Out[1]>>4)
                              );
                  curData[1]=(  ((curBase64Out[1]&15)<<4)
                              | (curBase64Out[2]>>2)
                              );
                  curData[2]=0;

                }
              else
                {
                  curData[0]=(  (curBase64Out[0]<<2)
                              | (curBase64Out[1]>>4)
                              );
                  curData[1]=(  ((curBase64Out[1]&15)<<4)
                              | (curBase64Out[2]>>2)
                              );
                  curData[2]=(  ((curBase64Out[2]&3)<<6)
                              | (curBase64Out[3])
                              );
                }
            }

          curBase64+=4;
          curData+=3;
        }
    }

    buff.data[lenNeeded-1]=0;
  }

#if 0
/* This code is dangerous--it prints out the cleartext password--so
 *  it should not be turned on just by DEBUG.
 */
#ifdef DEBUG
  printf("de64(%s): %s\n",
         base64,
         buff.data
         );
#endif
#endif

  return buff.data;
}

/* Returns an authentication realm for the user.  Currently hard-coded
 *  to return "roamd".
 */
const char* getRealm(const char* user)
{
  return "roamd";
}

/* Returns 1 if the credentials are acceptable, 0 if not, -1 on error.
 *  If user is non-NULL, then only that user has permission.
 */
int authorize(const char* user,
              const char* credential)
{
  /* Does the credential start with "Basic "? If not, it's an
   *  unimplemented authentication scheme.
   */
  const char* basicPrefix="Basic ";
  if (strstr(credential,basicPrefix)!=credential)
    {
#ifdef DEBUG
      fprintf(stderr,
              "authorize(\"%s\"): doesn't start with \"%s\"\n",
              credential,basicPrefix);
#endif
      return -1;
    }

  {
    /* OK, so what's basic-cookie? (see the BNF in RFC-2068, 11.1) */
    const char* basicCookie=credential+strlen(basicPrefix);
    if (!(*basicCookie))
      {
        fprintf(stderr,"authorize(): basicCookie==\"\"\n");
        return -1;
      }

    {
      /* OK, so what's user-pass? (see the BNF in RFC-2068, 11.1) */
      char* userColonPass=de64(basicCookie);
      if (!userColonPass)
        {
          fprintf(stderr,"authorize(): de64(basicCookie)==0\n");
          return -1;
        }

      {
        /* OK, so split it into user and password... */
        char* colon=strchr(userColonPass,':');
        if (!colon)
          {
            fprintf(stderr,"authorize(): user-pass does not contain a :\n");
            return -1;
          }
        *colon=0;

        /* Make sure it's the expected username */
        if (   user
            && strcmp(user,userColonPass)
            )
          {
#ifdef DEBUG
            fprintf(stderr,"authorize(): wrong username\n");
#endif
            return 0;
          }

        /* ...and have it checked */
        return authorizeBasic(userColonPass,colon+1);
      }
    }
  }
}

/* Returns 1 if the credentials are acceptable, 0 if not, -1 on error */
int authorizeBasic(const char* user,
                   const char* pwd)
{
  /* Look up the user */
  struct passwd* pwent=getpwnam(user);
  if (!pwent)
    {
#ifdef DEBUG
      fprintf(stderr,"authorizeBasic(): no such user as \"%s\"\n",user);
#endif
      return 0;
    }

  /* The first two characters of the stored encrypted password are
   *  the salt needed by crypt().
   */
  {
    char salt[3];

    salt[0]=pwent->pw_passwd[0];
    salt[1]=pwent->pw_passwd[1];
    salt[2]=0;

    {
      /* Encrypt it */
      const char* pwdCrypt=crypt(pwd,salt);
      if (!pwdCrypt)
        {
          fprintf(stderr,"authorizeBasic(): can't encrypt password\n");
          return -1;
        }
      
      /* Do the encrypted passwords match? */
      if (strcmp(pwdCrypt,pwent->pw_passwd)==0)
        return 1;
      else
        {
#ifdef DEBUG
          fprintf(stderr,"authorizeBasic(): passwords don't match\n");
#endif
          return 0;
        }
    }
  }      
}
