/*
 * The contents of this file are subject to the Mozilla Public License 
 * Version 1.0 (the "License"); you may not use this file except in compliance 
 * with the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/.
 * 
 * Software distributed under the License is distributed on an "AS IS" basis, 
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License 
 * for the specific language governing rights and limitations under the License. 
 * 
 * The Original Code is the roamd Version 0.1 code,
 * released on or about December 4, 1998.
 * 
 * The Initial Developer of the Original Code is John Richard Stracke,
 * Jr.  Portions created by John Richard Stracke, Jr. are Copyright (C)
 * John Richard Stracke, Jr. All Rights Reserved. 
 * 
 * Contributor(s): ______________________________________. 
 */
/*
 * Implements the initUsers() function from users.h
 *
 * $Revision: 1.8 $
 * $Author: francis $
 * $Date: 1999/03/21 01:52:57 $
 *
 * $Log: users.c,v $
 * Revision 1.8  1999/03/21 01:52:57  francis
 * Fixing some things from the last checkin:
 * In users.c, having eachPass() return N (this stuff was copied over
 * by hand from a version done on Solaris [where I had Purify], and I
 * left something out).  Also adding some debugs that helped me find
 * the problem.
 * In roamd.c, adding --noFork to the usage(); also fixing its procargs()
 * clause (before, it fell through to the "unrecognized option" part).
 * In Makefile, fixing the new purify part of clean.
 *
 * Revision 1.7  1999/03/21 01:23:59  francis
 * Fixing some bugs spotted by Purify, including a file descriptor leak.
 * Adding fork-per-connection (disable with the --noFork option).
 *
 * Revision 1.6  1998/12/13 06:44:50  francis
 * Adding includes to find malloc() on more platforms
 *
 * Revision 1.5  1998/12/06 20:02:22  francis
 * Adding ident lines
 *
 * Revision 1.4  1998/12/04 22:03:26  francis
 * Adding license terms to the source files.
 * Adding tryCrypt and tryCrypt.o to the clean target.
 * Deleting tryCrypt from the repository.
 * Adding Index.html and license.html.  Linking to license.html from
 * readme.html.  A bit of cosmetic work in readme.html.
 *
 * Revision 1.3  1998/12/04 20:02:41  francis
 * Fixing initUsers() to be cheap to call a second time
 *
 * Revision 1.2  1998/12/04 20:01:03  francis
 * Adding documentation headers and comments
 *
 */
#include "ident.h"
IDENTIFY(ROAMD_USERS_C,"$Id: users.c,v 1.8 1999/03/21 01:52:57 francis Exp $")

#include <sys/types.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <pwd.h>
#include <malloc.h>
#include <stdlib.h>

#include "users.h"

/*
 * Runs through the directory, looking for directories whose names are
 *  legal usernames.  If res==0, just counts them; otherwise, copies
 *  them into res.  This avoids copying code.
 *
 * Returns <0 on error.  Otherwise, if res==0, returns the number of
 *  directories found; if res!=0, returns the number stored in res
 *  (which may not be the same as the count--say, if strdup() fails).
 */
static int eachPass(const char* userDir,
                    const char** res,
                    int count)
{
  struct dirent* cur;
  int N=0;

  DIR* dir;

#ifdef DEBUG
  fprintf(stderr,"eachPass(%s)\n",userDir);
#endif

  dir=opendir(userDir);
  if (!dir)
    {
      perror("opendir");
      return -1;
    }

  while (cur=readdir(dir))
    {
      char path[MAXPATHLEN];
      struct stat buf;
  
      const char* f=cur->d_name;
      sprintf(path,"%s/%s",userDir,f);
      if (stat(path,&buf)<0)
        continue;
      if (!S_ISDIR(buf.st_mode))
        continue;

      if (!getpwnam(f))
        {
#ifdef DEBUG
          fprintf(stderr,"User %s can't be used: getpwnam: %s\n",
                  f,
                  strerror(errno));
#endif
          continue;
        }

      if (res)
        {
          res[N]=strdup(f);
          if (res[N])
            {
              N++;
              if (N>=count)
                break;
            }
        }
      else
        N++;
    }

  if (res)
    res[N]=0;

  closedir(dir);
  return N;
}

/*
 * Given the directory under which user profiles are stored (in the default
 *  configuration, "/etc/roamd/users"), returns a 0-terminated array of
 *  usernames.  The array is terminated by 0.  If this function is
 *  called twice, the second time it will return the same array without
 *  rebuilding it.
 */
const char** initUsers(const char* userDir)
{
  static const char** users=0;
  int N;

  if (users)
    return users;

  /* Count the users */
  N=eachPass(userDir,0,0);
  if (N<0)
    return 0;

#ifdef DEBUG
  fprintf(stderr,"initUsers(): eachPass() counted %d users\n",N);
#endif

  {
    /* Allocate the array to hold them.  Note that it's N+1 instead of N,
     *  because it's 0-terminated.
     */
    const char** res=(const char**)(malloc((N+1)*sizeof(const char*)));
    if (!res)
      return 0;

    /* Copy the usernames into the array */
    if (eachPass(userDir,res,N)<0)
      {
        int tmp=errno;
        free(res);
        errno=tmp;
        return 0;
      }

#ifdef DEBUG
    {
      int i;
      fprintf(stderr,"users=={");
      for (i=0; res[i]; i++)
        {
          if (i>0)
            fprintf(stderr,",");
          fprintf(stderr,"%s",res[i]);
        }
      fprintf(stderr,"}\n");
    }
#endif

    return users=res;
  }
}
